/*
 * my_SD.c
 *
 * Created: 2/24/2012 7:55:10 PM
 *  Author: Umair Mukati
 */ 
#define F_CPU 8000000UL
#include <avr/io.h>
#include "my_SD.h"
#include "SPI_lib.h"
#include "usart.h"
#include <util/delay.h>

/*
unsigned char SD_init(void)
{
	unsigned char i=0, response=0;
	unsigned int retry=0;
	SD_version = 0;
	
	usart_init(UART_BAUD_SELECT(9600,F_CPU),1,1);
	spi_init();
	
	usart_puts("\n\r1");
	
	CS_DEASSERT;
	
	/ *First 74 Clocks* /
	for(i=0;i<10;i++)
		SPI_rec();
	
	/ *Software Reset* /
	CS_ASSERT;
	for(retry=0;response != 0x01;++retry)
	{
		response = SD_sendCommand(GO_IDLE_STATE, 0); //send 'reset & go idle' command
		if(retry == 0x1ff){
			CS_DEASSERT;
			usart_puts("\r\nCard not detected!");
   			return 1;   //time out, card not detected
		}
	}
	CS_DEASSERT;
	usart_puts("\r\nCard detected!");
	retry = 0;
	CS_ASSERT;
	
	/ *Check power supply status, mendatory for SDHC card* /
	do
	{
		response = SD_sendCommand(SEND_IF_COND,0x000001AA);
		retry++;
		if(retry>0x7ff) 
		{
			SD_version = 1;
			break;
			
		} //time out
		
	}while(response != 0x01);
	
	CS_DEASSERT;
	retry = 0;
	CS_ASSERT;
	
	if(SD_version == 1){
		do
		{
			/ * CMD55, must be sent before sending any ACMD command * /
			response = SD_sendCommand(APP_CMD,0);
			/ *ACMD41* /
			response = SD_sendCommand(SD_SEND_OP_COND,0x00000000);
			retry++;
			if(retry>0xfe) 
			{
				if (response != 0x01)
				{
					/ * It can be an MMC or SDHC * /
					SD_version = 0;
					retry = 0;
					do
					{
						/ * Checking for MMC* /
						/ *CMD55, must be sent before sending any ACMD command* /
						response = SD_sendCommand(APP_CMD,0);
						response = SD_sendCommand(SEND_OP_COND,0x00000000); //ACMD41
						retry++;
						if(retry>0xfe) 
						{
							/ * It can be SDHC * /
							SD_version = 2;
							break;
						} 
					}while(response != 0x00);
					break;
				}
			} 
		}while(response != 0x00);
	}
	
	retry = 0;
	
	if(SD_version == 2)
	{
		do
		{
			response = SD_sendCommand(APP_CMD,0); //CMD55, must be sent before sending any ACMD command
			response = SD_sendCommand(SD_SEND_OP_COND,0x40000000); //ACMD41
			retry++;
			if(retry>0x7ff) 
			{
				return 2;  //time out, card initialization failed
			} 
		}while(response != 0x00);
		do
		{
			response = SD_sendCommand(READ_OCR,0);
			retry++;
			if(retry>0x7ff) 
			{
				cardType = 0;
				break;
			} //time out

		}while(response != 0x00);
	}

	retry = 0;
	
	CS_DEASSERT;
	
	switch(SD_version)
	{
		case 0:
			usart_puts("\r\nMMC CARD");
			break;
		case 1:
			usart_puts("\r\nSD Version 1");
			break;
		case 2:
			usart_puts("\r\nSD Version 2");
			break;
	}

	CS_ASSERT;
	
	if(SD_sendCommand(SET_BLOCK_LEN, 512)); //set block size to 512; default size is 512

	CS_DEASSERT;
	
	usart_puts("\r\nALHAMDULILLAH\r\n");
	SPI_HIGH_SPEED;
	return 0; //successful return
}*/
						

uint8_t SD_init()
{
    CS_DEASSERT;

    /* initialize SPI with lowest frequency; max. 400kHz during identification mode of card */
    SPCR = (0 << SPIE) | /* SPI Interrupt Enable */
           (1 << SPE)  | /* SPI Enable */
           (0 << DORD) | /* Data Order: MSB first */
           (1 << MSTR) | /* Master mode */
           (0 << CPOL) | /* Clock Polarity: SCK low when idle */
           (0 << CPHA) | /* Clock Phase: sample on rising SCK edge */
           (1 << SPR1) | /* Clock Frequency: f_OSC / 128 */
           (1 << SPR0);
    SPSR &= ~(1 << SPI2X); /* No doubled clock frequency */

	usart_puts("1");
    /* card needs 74 cycles minimum to start up */
    for(uint8_t i = 0; i < 10; ++i)
    {
        /* wait 8 clock cycles */
        SPI_rec();
    }
	usart_puts("2");
    /* address card */
    CS_ASSERT;

    /* reset card */
    uint8_t response;
    for(uint16_t i = 0; ; ++i)
    {
        response = SD_sendCommand(GO_IDLE_STATE, 0);
        if(response == (1 << 0))
            break;		
        if(i == 0x1ff)
        {
            CS_DEASSERT;
			usart_puts("Not Detected");
            return 0;
        }
    }
	usart_puts("Pass");
    
    /* wait for card to get ready */
    for(uint16_t i = 0; ; ++i)
    {
        response = SD_sendCommand(SEND_OP_COND, 0);
        if(!(response & (1 << 0)))
            break;

        if(i == 0x7fff)
        {
            CS_DEASSERT;
            return 0;
        }
    }

    /* set block size to 512 bytes */
    if(SD_sendCommand(SET_BLOCK_LEN, 512))
    {
        CS_DEASSERT;
        return 0;
    }

    /* deaddress card */
    CS_DEASSERT;

    /* switch to highest SPI frequency possible */
    SPCR &= ~((1 << SPR1) | (1 << SPR0)); /* Clock Frequency: f_OSC / 4 */
    SPSR |= (1 << SPI2X); /* Doubled Clock Frequency: f_OSC / 2 */

	return 1;
}






unsigned char SD_sendCommand(unsigned char cmd, unsigned long arg)
{
	unsigned char response, retry=0, status;
	//SD card accepts byte address while SDHC accepts block address in multiples of 512
	//so, if it's SD card we need to convert block address into corresponding byte address by 
	//multipying it with 512. which is equivalent to shifting it left 9 times
	//following 'if' loop does that
	SDHC_flag = 0;
	if(SDHC_flag == 0)		
		if(cmd == READ_SINGLE_BLOCK  ||
		cmd == READ_MULTIPLE_BLOCKS  ||
		cmd == WRITE_SINGLE_BLOCK    ||
		cmd == WRITE_MULTIPLE_BLOCKS ||
		cmd == ERASE_BLOCK_START_ADDR|| 
		cmd == ERASE_BLOCK_END_ADDR ) 
		{
			arg = arg << 9;
		}	   

	//CS_ASSERT;
	SPI_send(cmd|0x40); //send command, first two bits always '01'
	SPI_send(arg>>24);
	SPI_send(arg>>16);
	SPI_send(arg>>8);
	SPI_send(arg);
		
	if(cmd == SEND_IF_COND)	 //it is compulsory to send correct CRC for CMD8 (CRC=0x87) & CMD0 (CRC=0x95)
		SPI_send(0x87);    //for remaining commands, CRC is ignored in SPI mode
	else 
		SPI_send(0xFF); 
		
	while((response = SPI_rec()) == 0xff) //wait response
		if(retry++ > 0xfe) break; //time out error
		
	if(response == 0x00 && cmd == 58)  //checking response of CMD58
	{
		status = SPI_rec() & 0x40;     //first byte of the OCR register (bit 31:24)
		if(status == 0x40) SDHC_flag = 1;  //we need it to verify SDHC card
		else SDHC_flag = 0;
		SPI_rec(); //remaining 3 bytes of the OCR register are ignored here
		SPI_rec(); //one can use these bytes to check power supply limits of SD
		SPI_rec(); 
	}
		
	SPI_rec(); //extra 8 CLK
	//if((cmd != WRITE_SINGLE_BLOCK) && (cmd != READ_SINGLE_BLOCK))CS_DEASSERT;
	
	
	return response; //return state
}

unsigned char SD_erase (unsigned long startBlock, unsigned long totalBlocks)
{
	unsigned char response;
	response = SD_sendCommand(ERASE_BLOCK_START_ADDR, startBlock); //send starting block address
	if(response != 0x00) //check for SD status: 0x00 - OK (No flags set)
		return response;
	response = SD_sendCommand(ERASE_BLOCK_END_ADDR,(startBlock + totalBlocks - 1)); //send end block address
	if(response != 0x00)
		return response;
	response = SD_sendCommand(ERASE_SELECTED_BLOCKS, 0); //erase all selected blocks
	if(response != 0x00)
		return response;
	return 0; //normal return
}

//******************************************************************
//Function	: to read a single block from SD card
//Arguments	: none
//return	: unsigned char; will be 0 if no error,
// 			  otherwise the response byte will be sent
//******************************************************************
unsigned char SD_readSingleBlock(unsigned long startBlock)
{
	unsigned char response;
	unsigned int i, retry=0;
	response = SD_sendCommand(READ_SINGLE_BLOCK, startBlock); //read a Block command
	if(response != 0x00){usart_puts("Notdetected!\r\n"); return response;} //check for SD status: 0x00 - OK (No flags set)
	//CS_ASSERT;
	retry = 0;
	while(SPI_rec() != 0xfe) //wait for start block token 0xfe (0x11111110)
	if(retry++ > 0xfffe){CS_DEASSERT; return 1;} //return if time-out
	
	for(i=0; i<512; i++) //read 512 bytes
		buffer[i] = SPI_rec();
	SPI_rec(); //receive incoming CRC (16-bit), CRC is ignored here
	SPI_rec();
	SPI_rec(); //extra 8 clock pulses
	CS_DEASSERT;
	usart_puts("\rEnds!");
	return 0;
}

//******************************************************************
//Function	: to write to a single block of SD card
//Arguments	: none
//return	: unsigned char; will be 0 if no error,
// 			  otherwise the response byte will be sent
//******************************************************************
unsigned char SD_writeSingleBlock(unsigned long startBlock)
{
	unsigned char response;
	unsigned int i, retry=0;
	usart_puts("\r\n");
	usart_puts("1\n\r");
	
	CS_ASSERT;
	response = SD_sendCommand(WRITE_SINGLE_BLOCK, startBlock); //write a Block command
	usart_puts("2\n\r");
	if(response != 0x00) return response; //check for SD status: 0x00 - OK (No flags set)	
	usart_puts("3\n\r");
	SPI_send(0xfe);     //Send start block token 0xfe (0x11111110)
	for(i=0; i<512; i++)    //send 512 bytes data
		SPI_send(buffer[i]);
	SPI_send(0xff);     //transmit dummy CRC (16-bit), CRC is ignored here
	SPI_send(0xff);
	response = SPI_rec();
	if( (response & 0x0f) != 0x05) //response= 0xXXX0AAA1 ; AAA='010' - data accepted
	{                              //AAA='101'-data rejected due to CRC error
		CS_DEASSERT;              //AAA='110'-data rejected due to write error
		usart_puts("ERROR");
		return response;
	}
	while(!SPI_rec()) //wait for SD card to complete writing and get idle
		if(retry++ > 0xfffe){CS_DEASSERT;usart_puts("4\n\r"); return 1;}
	CS_DEASSERT;
	SPI_send(0xff);   //just spend 8 clock cycle delay before reasserting the CS line
	CS_ASSERT;         //re-asserting the CS line to verify if card is still busy
	while(!SPI_rec()) //wait for SD card to complete writing and get idle
		if(retry++ > 0xfffe){CS_DEASSERT; usart_puts("5\n\r");return 1;}
	CS_DEASSERT;
	usart_puts("\rDONE SAVING");
	return 0;
}